"""Abstract base class for continuation engines."""

from hiten.algorithms.types.core import _HitenBaseEngine


class _ContinuationEngine(_HitenBaseEngine):
    """Shared base class for continuation engines."""
