"""Define the base class for correction engines.

This module provides the base class for correction engines.
"""

from hiten.algorithms.types.core import _HitenBaseEngine


class _CorrectionEngine(_HitenBaseEngine):
    """Shared base class for correction engines."""